/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package agsecres.tool;

import agsecres.helper.*;
import java.lang.*;
import java.io.*;
import java.util.*;
import java.text.*;

public class Interpreter {
 static int last_menu=-1;

 public static final CmdDesc base_cmd_table[]={
  new CmdDesc("exit"       , null           ,  null                                      , new CMD_EXIT()),
  new CmdDesc("help"       , "S"            ,  "[cmd]"                                   , new CMD_HELP()),

  new CmdDesc("output"     , "S"            ,  "console|filename"                        , new CMD_OUTPUT()),

  new CmdDesc("var"        , "SS"           ,  "varname str|int"                         , new CMD_VAR()),
  new CmdDesc("set"        , "SS"           ,  "[varname|varname value]"                 , new CMD_SET()),

  new CmdDesc("echo"       , "S"            ,  "on|off"                                  , new CMD_ECHO()),
  new CmdDesc("comm"       , "S"            ,  "on|off"                                  , new CMD_COMM()),
  new CmdDesc("print"      , "SSSSSSSSSSSSS",  "arg1 arg2 ... argn"                      , new CMD_PRINT()),
  new CmdDesc("run"        , "SSSSSSSSSSSSS",  "arg1 arg2 ... argn"                      , new CMD_RUN()),
  new CmdDesc("assert"     , "aIvSeS"       ,  "-a argnum|-v varname [-e errormsg]"      , new CMD_ASSERT()),

  new CmdDesc("mod"        , "S"            ,  "module"                                  , new CMD_MOD()),

  new CmdDesc("test"       , "I"            ,  "arg"                                     , new CMD_TEST())
 };

 public static CmdDesc[] cmd_table() {
  CmdDesc[] mod_table=ModRegistry.cmd_table();

  int cmdnum=base_cmd_table.length+mod_table.length;

  CmdDesc cmd_table[]=new CmdDesc[cmdnum];

  System.arraycopy(base_cmd_table,0,cmd_table,0,base_cmd_table.length);
  System.arraycopy(mod_table,0,cmd_table,base_cmd_table.length,mod_table.length);

  return cmd_table;
 }

 public static CmdDesc lookup_cmd(String cmd) {
  for(int i=0;i<cmd_table().length;i++) {
   CmdDesc desc=cmd_table()[i];

   if (desc.name.equals(cmd)) return desc;
  }

  return null;
 }

 public static boolean run_cmd(Shell.Cmd cmd) throws Throwable {
  CmdDesc desc=cmd.desc;
  CmdHandler handler=desc.handler;

  return handler.run(cmd.options);
 }

 //shell commands implementation
 public static class CMD_EXIT implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   Shell.stop();

   return true;
  }
 }

 public static class CMD_HELP implements CmdHandler {
  private static void print_usage(CmdDesc cmd) {
   String usage="";

   if (cmd.usage!=null) usage+=cmd.usage;

   Shell.println(Utils.padded_string(cmd.name,0x10)+"   "+usage);
  }

  public boolean run(Shell.Option[] options) throws Throwable {
   String cmdname=null;

   if (options.length<=1) {
    if (options.length==1) {
     if (options[0].pure_arg()) {
      cmdname=options[0].str_arg();
     }
    }

    if (cmdname!=null) {
     CmdDesc cmd=lookup_cmd(cmdname);

     if (cmd!=null) {
      print_usage(cmd);
     } else {
      Shell.err_string="unknown command";
      return false;
     }
    } else {
     for(int i=0;i<cmd_table().length;i++) {
      CmdDesc cmd=cmd_table()[i];

      print_usage(cmd);
     }
    }
   } else {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   return true;
  }
 }

 public static class CMD_OUTPUT implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String name=null;

   if (options.length!=0) {
    if (options[0].pure_arg()) {
     name=options[0].str_arg();
    }
   }

   if (name==null) {
    Shell.err_string="missing output name argument";
    return false;
   }

   if (name.equals("console")) {
    if (Shell.getOutput()!=Shell.CONSOLE_OUTPUT) {
     Shell.closeOutput();
    }

    Shell.setOutput(Shell.CONSOLE_OUTPUT);
   } else {
    try {
     Shell.setOutput(new PrintWriter(new FileOutputStream(name,false),true,Shell.CHARSET));
    } catch(Throwable t) {
     Shell.setOutput(Shell.CONSOLE_OUTPUT);
     Shell.err_string="invalid output name";
     return false;
    }
   }
 
   return true;
  }
 }

 public static class CMD_VAR implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String name=null;
   String type=null;


   if (options.length==2) {
    if (options[0].pure_arg()) {
     name=options[0].str_arg();
    }

    if (name==null) {
     Shell.err_string="missing name argument";
     return false;
    }

    if (options[1].pure_arg()) {
     type=options[1].str_arg();
    }

    if (type==null) {
     Shell.err_string="missing type argument";
     return false;
    }

    Vars.declare(name,type);
   } else {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   return true;
  }
 }

 public static class CMD_SET implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String name=null;
   String val=null;

   if (options.length==0) {
    Vars.print();
   } else

   if (options.length==1) {
    //clear val
    if (options[0].pure_arg()) {
     name=options[0].str_arg();
    }

    Vars.clear(name);
   } else

   if (options.length==2) {
    //set val
    if (options[0].pure_arg()) {
     name=options[0].str_arg();
    }

    if (options[1].pure_arg()) {
     val=options[1].str_arg();
    }

    Vars.Var v=Vars.get_var(name);

    if (v!=null) {
     Vars.Proto proto=v.proto();
 
     switch(proto.type()) {
      case Vars.VAR_INT:
       Integer ival=Shell.parse_hex_value(val);
      
       if (ival==null) {
        Shell.err_string="invalid integer val";
        return false;
       }
 
       Vars.set(name,ival.intValue());
       break;
      case Vars.VAR_STR:
       Vars.set(name,val);
       break;
     }
    } else {
     Shell.err_string="invalid number of arguments";
     return false;
    }
   } else {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   return true;
  }
 }
 
 public static class CMD_ECHO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length==0) {
    String echo="";
 
    if (Shell.echo()) {
     echo="on";
    } else {
     echo="off";
    }
 
    Shell.println("echo is "+echo);
   } else

   if (options.length==1) {
    String echo=options[0].str_arg();

    if (echo.equals("on")) {
     Shell.echo(true);
    } else

    if (echo.equals("off")) {
     Shell.echo(false);
    } else {
     Shell.err_string="invalid srgument";
     return false;
    }
   } else {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   return true;
  }
 }

 public static class CMD_COMM implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length==0) {
    String comm="";
 
    if (Shell.comm()) {
     comm="on";
    } else {
     comm="off";
    }
 
    Shell.println("comm is "+comm);
   } else

   if (options.length==1) {
    String comm=options[0].str_arg();

    if (comm.equals("on")) {
     Shell.comm(true);
    } else

    if (comm.equals("off")) {
     Shell.comm(false);
    } else {
     Shell.err_string="invalid argument";
     return false;
    }
   } else {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   return true;
  }
 }

 public static class CMD_PRINT implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   for(int i=0;i<options.length;i++) {
    Shell.print(options[i].str_arg());
   }

   if (options.length>0) Shell.println("");

   return true;
  }
 }

 public static class CMD_RUN implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String filename=null;

   if (options.length<1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   if (options[0].pure_arg()) {
    filename=options[0].str_arg();
   }

   if (filename==null) {
    Shell.err_string="missing filename argument";
    return false;
   }

   Shell.ScriptArgs args=new Shell.ScriptArgs(options);
   Shell.run_script(filename,args);

   return true;
  }
 }

 public static class CMD_ASSERT implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   int argnum=-1;
   String varname=null;
   String err=null;

   Shell.Option opt=Shell.getopt(options,'a');

   if (opt!=null) {
    argnum=opt.int_arg();
   } 

   opt=Shell.getopt(options,'v');

   if (opt!=null) {
    varname=opt.str_arg();
   } 

   opt=Shell.getopt(options,'e');

   if (opt!=null) {
    err=opt.str_arg();
   } 

   if (argnum!=-1) {
    if (Shell.argnum()!=argnum) {
     if (err!=null) { 
      Shell.println(err);
     }

     Shell.assert_status(false);
    }
   } else

   if (varname!=null) {
    Vars.Var var=Vars.get_var(varname);

    if ((var==null)||(var.val()==null)) {
     if (err!=null) { 
      Shell.println(err);
     }

     Shell.assert_status(false);
    }
   }

   return true;
  }
 }

 public static class CMD_MOD implements CmdHandler {
  private static String mod_info(TolkitMod mod) {
   return Utils.padded_string(mod.name(),8)+" ["+mod.desc()+"]";
  }

  public boolean run(Shell.Option[] options) throws Throwable {
   String modname=null;

   if (options.length==0) {
    for(int i=0;i<ModRegistry.size();i++) {
     TolkitMod mod=ModRegistry.get(i);

     String info=mod_info(mod)+" "+mod.cmd_table().length;
     Shell.println("- "+info);
    }
   } else

   if (options.length==1) {
    if (options[0].pure_arg()) {
     modname=options[0].str_arg();
    }
   } else {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   if (modname!=null) {
    if (ModRegistry.is_loaded(modname)) {
     Shell.err_string="module already loaded";
     return false;     
    }

    if (!ModRegistry.load_module(modname)) {
     Shell.err_string="failed to load module: "+modname;
     return false;     
    }

    TolkitMod mod=ModRegistry.get(modname);

    Shell.println("loaded "+mod_info(mod));
   }

   return true;
  }
 }

 public static class CMD_TEST implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   Shell.println("TEST command");

   return true;
  }
 }
}