/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package agsecres.helper;

import agsecres.tool.*;
import java.lang.*;
import java.lang.reflect.*;
import java.util.*;
import java.io.*;

public class ELF {
 public static class Elf32_Ehdr {
  int magic;
  int ident1;
  int ident2;
  int ident3;
  short e_type;
  short e_machine;
  int e_version;
  int e_entry;
  int e_phoff;
  int e_shoff;
  int e_flags;
  short e_ehsize;
  short e_phentsize;
  short e_phnum;
  short e_shentsize;
  short e_shnum;
  short e_shstrndx;
 };

 public static class Elf32_Shdr {
  int sh_name;
  int sh_type;
  int sh_flags;
  int sh_addr;
  int sh_offset;
  int sh_size;
  int sh_link;
  int sh_info;
  int sh_addralign;
  int sh_entsize;
 };

 public static class Elf32_Symbol {
  int st_name;
  int st_value;
  int st_size;
  byte st_info;
  byte st_other;
  short st_shndx;
 };

 public static class Section {
  ELF elf;
  Elf32_Shdr shdr;
  String name;

  public Section(ELF elf,Elf32_Shdr shdr) {
   this.elf=elf;
   this.shdr=shdr;
  }

  public byte[] read(int off,int len) {
   return elf.read_n(shdr.sh_offset+off,len);
  }  

  public int string_table_off() {
   return elf.string_table_off();
  }

  public String name() {
   if (name==null) {
    name=elf.read_string(elf.string_table_off()+shdr.sh_name,0x40);
   }

   return name;
  }

  public int type() {
   return shdr.sh_type;
  }

  public int offset() {
   return shdr.sh_offset;
  }

  public int addr() {
   return shdr.sh_addr;
  }

  public int size() {
   return shdr.sh_size;
  }

  public int entsize() {
   return shdr.sh_entsize;
  }

  public void print() {
   elf.print_instance(shdr);
  }
 }

 public static class Symbol {
  ELF elf;
  Elf32_Symbol sym;
  String name;

  public Symbol(ELF elf,Elf32_Symbol sym) {
   this.elf=elf;
   this.sym=sym;
  }

  public String name() {
   if (name==null) {
    name=elf.read_string(elf.dynstr_off()+sym.st_name,0x40);
   }

   return name;
  }

  public int value() {
   Section sec=elf.sections[sym.st_shndx];

   return sec.offset()+sym.st_value-sec.addr();
  }

  public int size() {
   return sym.st_size;
  }

  public void print() {
   PaddedPrinter pp=Shell.get_pp();

   pp.println("["+name()+"]");
   pp.pad(2,"");

   pp.println("val:  "+Utils.hex_value(sym.st_value,8));
   pp.println("size: "+Utils.hex_value(sym.st_size,8));

   pp.leave();
  }
 }

 public static final int ELF_MAGIC = 0x464c457f;

 public static final int ET_REL  = 1; 
 public static final int ET_EXEC = 2;
 public static final int ET_DYN  = 3;

 public static final int PT_LOAD        = 1;
 public static final int PT_DYNAMIC     = 2;

 public static final int SHT_NULL         = 0x0;
 public static final int SHT_PROGBITS     = 0x1;
 public static final int SHT_SYMTAB       = 0x2;
 public static final int SHT_STRTAB       = 0x3;
 public static final int SHT_RELA         = 0x4;
 public static final int SHT_HASH         = 0x5;
 public static final int SHT_DYNAMIC      = 0x6;
 public static final int SHT_NOTE         = 0x7;
 public static final int SHT_NOBITS       = 0x8;
 public static final int SHT_REL          = 0x9;
 public static final int SHT_SHLIB        = 0xA;
 public static final int SHT_DYNSYM        = 0xB;
 public static final int SHT_INIT_ARRAY    = 0xE;
 public static final int SHT_FINI_ARRAY    = 0xF;
 public static final int SHT_PREINIT_ARRAY = 0x10;
 public static final int SHT_GROUP         = 0x11;
 public static final int SHT_SYMTAB_SHNDX  = 0x12;
 public static final int SHT_LOOS          = 0x60000000;
 public static final int SHT_HIOS          = 0x6FFFFFFF;
 public static final int SHT_GNU_LIBLIST   = 0x6FFFFFF7;
 public static final int SHT_LOPROC        = 0x70000000;
 public static final int SHT_HIPROC        = 0x7FFFFFFF;
 public static final int SHT_LOUSER        = 0x80000000;
 public static final int SHT_HIUSER        = 0xFFFFFFFF;

 byte data[];
 Hashtable<String,Symbol> syms;
 ByteInput bi;

 Elf32_Ehdr hdr;
 Section sections[];
 Symbol symbols[];

 int dynstr_off;

 public static ELF load(String filename) {
  if (Utils.file_exists(filename)) {
   byte data[]=Utils.load_file(filename);

   return new ELF(data);
  }

  return null;
 }

 public byte read_byte(int off) {
  bi.set_pos(off);

  return bi.read_1();
 }

 public short read_word(int off) {
  bi.set_pos(off);

  return bi.read_2();
 }

 public int read_dword(int off) {
  bi.set_pos(off);

  return bi.read_4();
 }

 public String read_string(int off,int max) {
  int len=0;

  for(int i=0;i<max;i++) {
   if (read_byte(off+i)!=0) len++;
    else break;
  }  

  byte data[]=new byte[len];


  for(int i=0;i<len;i++) {
   data[i]=(byte)read_byte(off+i);
  }

  return new String(data);
 }

 public byte[] read_n(int off,int len) {
  bi.set_pos(off);

  return bi.read_n(len);
 }

 public void fill_instance(Object instance,Class clazz,int off) {
  try {
   Field ftab[]=clazz.getDeclaredFields();

   for(int i=0;i<ftab.length;i++) {
    Field f=ftab[i];
    f.setAccessible(true);

    String type=f.getType().getName();

    if (type.equals("byte")) {
     byte val=read_byte(off);

     f.set(instance,new Byte(val));

     off+=1;
    } else

    if (type.equals("short")) {
     short val=read_word(off);

     f.set(instance,new Short(val));

     off+=2;
    } else

    if (type.equals("int")) {
     int val=read_dword(off);

     f.set(instance,new Integer(val));

     off+=4;
    } else break;
   }
  } catch(Throwable t) {
   t.printStackTrace();
  }
 }

 public void print_instance(Object instance) {
  try {
   PaddedPrinter pp=Shell.get_pp();

   pp.println("["+instance.getClass().getName()+"]");
   pp.pad(2,"");

   Class clazz=instance.getClass();

   Field ftab[]=clazz.getDeclaredFields();

   for(int i=0;i<ftab.length;i++) {
    Field f=ftab[i];

    String name=f.getName();

    String type=f.getType().getName();

    int val=-1;

    if (type.equals("byte")) {
     Byte bval=(Byte)f.get(instance);
     val=((byte)bval.byteValue())&0xff;
    } else

    if (type.equals("short")) {
     Short sval=(Short)f.get(instance);
     val=((int)sval.shortValue())&0xffff;
    } else

    if (type.equals("int")) {
     Integer ival=(Integer)f.get(instance);
     val=ival.intValue();
    }

    pp.println(Utils.padded_string(name,16)+":  "+Utils.hex_value(val,8));
   }

   pp.leave();
  } catch(Throwable t) {
   t.printStackTrace();
  }
 }

 public Object read_instance(Class clazz,int off) {
  Object instance=null;

  try {
   instance=clazz.newInstance();

   fill_instance(instance,clazz,off);
  } catch(Throwable t) {
   instance=null;
   t.printStackTrace();
  }
   
  return instance;
 }

 public int string_table_off() {
  return sections[hdr.e_shstrndx].shdr.sh_offset;
 }

 public int dynstr_off() {
  if (dynstr_off==0) {
   Section sec=lookup_section(".dynstr");

   if (sec!=null) {
    dynstr_off=sec.offset();
   }
  }

  return dynstr_off;
 }

 private void read_header() throws Throwable {
  hdr=(Elf32_Ehdr)read_instance(Elf32_Ehdr.class,0);
 }

 private void read_sections() throws Throwable {
  sections=new Section[hdr.e_shnum];

  int off=hdr.e_shoff;

  for(int i=0;i<sections.length;i++) {
   Elf32_Shdr shdr=(Elf32_Shdr)read_instance(Elf32_Shdr.class,off);

   sections[i]=new Section(this,shdr);
   off+=hdr.e_shentsize;
  }
 }

 private void load_dynamic_symbols() throws Throwable {
  Section dynsym=null;

  for(int i=0;i<sections.length;i++) {
   String name=sections[i].name();
   Section section=sections[i];

   if (section.type()==SHT_DYNSYM) {
    dynsym=section;
    break;     
   }
  } 

  if (dynsym!=null) {
   int off=dynsym.offset();

   int cnt=dynsym.size()/dynsym.entsize();

   symbols=new Symbol[cnt];

   for(int i=0;i<cnt;i++) {
    Elf32_Symbol sym=(Elf32_Symbol)read_instance(Elf32_Symbol.class,off);

    Symbol symbol=new Symbol(this,sym);
    symbols[i]=symbol;

    syms.put(symbol.name(),symbol);

    off+=dynsym.entsize();
   }
  }     
 }

 public Section lookup_section(String name) {
  Section res=null;

  for(int i=0;i<sections.length;i++) {
   Section s=sections[i];

   if (s.name().equals(name)) return s;
  }

  return res;
 }

 public Symbol lookup_symbol(String name) {
  return syms.get(name);
 }

 private void parse_exec_image() {
  try {
   read_header();
   read_sections();
   load_dynamic_symbols();   
  } catch(Throwable t) {
   t.printStackTrace();
  }
 }

 public ELF(byte data[]) {
  this.data=data;
  syms=new Hashtable<String,Symbol>();

  bi=new ByteInput(data);
  bi.little_endian();

  int magic=read_dword(0x00);

  if (magic==ELF_MAGIC) {
   int type=read_word(0x10);

   if (type==ET_DYN) parse_exec_image();
  }
 }
}
