/*## (c) SECURITY EXPLORATIONS    2018 Poland                                #*/
/*##     http://www.security-explorations.com                                #*/
/*## (c) AG Security Research     2019-2022 Poland                           #*/
/*##     http://www.agsecurityresearch.com                                   #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

package mod.mspr;

import agsecres.tool.*;
import agsecres.helper.*;
import mod.cdn.*;
import mod.vod.*;
import java.lang.*;
import java.util.*;
import java.io.*;
import java.math.*;
import java.util.concurrent.*;

import java.awt.*;
import java.awt.image.*;
import javax.imageio.*;
import java.nio.file.*;

public class Init implements TolkitMod {
 public static final String DEVROOTKEY_DIR = "devrootkey";
 public static final String BASE_DIR       = "secrets"+File.separatorChar+DEVROOTKEY_DIR;

 static {
  Utils.mkdir("samples");
  Utils.mkdir("content");
 }

 public static final CmdDesc cmd_table[]={
  new CmdDesc("device"         , null           ,  null                                                         , new CMD_DEVICE()),
  new CmdDesc("grpkey"         , null           ,  null                                                         , new CMD_GRPKEY()),
  new CmdDesc("grpcert"        , null           ,  null                                                         , new CMD_GRPCERT()),
  new CmdDesc("devcert"        , null           ,  null                                                         , new CMD_DEVCERT()),
  new CmdDesc("devchain"       , null           ,  null                                                         , new CMD_DEVCHAIN()),

  new CmdDesc("maninfo"        , "AS"           ,  "[-A assetid]"                                               , new CMD_MANINFO()),
  new CmdDesc("urlinfo"        , "AS"           ,  "[-A assetid]"                                               , new CMD_URLINFO()),

  new CmdDesc("getlicense"     , "AS"           ,  "[-A assetid]"                                               , new CMD_GETLICENSE()),
  new CmdDesc("makemp4"        , "ASaSvStS"     ,  "[-A assetid] -a audiostream -v videostream -t timedesc"     , new CMD_MAKEMP4()),
  new CmdDesc("downmp4"        , "ASaSvStS"     ,  "[-A assetid] -a audiostream -v videostream -t timedesc"     , new CMD_DOWNMP4()),

  new CmdDesc("assetinfo"      , "AS"           ,  "[-A assetid]"                                               , new CMD_ASSETINFO()),

  new CmdDesc("mp4dump"        , "fSbSv"        ,  "[-f file][-v][-b path]"                                     , new CMD_MP4DUMP()),

  new CmdDesc("secrets"        , null           ,  null                                                         , new CMD_SECRETS()),
  new CmdDesc("bcertinfo"      , "S"            ,  "bcertfile"                                                  , new CMD_BCERTINFO()),
  new CmdDesc("checkwatermark" , "ASiIqS"       ,  "[-A assetid][-i fragment_idx][-q quality]"                  , new CMD_CHECKWATERMARK()),

  new CmdDesc("extractsecrets" , "S"            ,  "playreadybinary"                                            , new CMD_EXTRACTSECRETS()),

  new CmdDesc("gentest"        , null           ,  null                                                         , new CMD_GENTEST()),
  new CmdDesc("liccrawl"       , "SI"           ,  "collection cnt"                                             , new CMD_LICCRAWL()),
 };

 public String name() {
  return "mspr";
 }

 public String desc() {
  return "MS Play Ready toolkit";
 }

 public CmdDesc[] cmd_table() {
  return cmd_table;
 }

 private static String parse_assetid(Shell.Option[] options) {
  String assetid=null;

  Shell.Option opt=Shell.getopt(options,'A');

  if (opt!=null) {
   assetid=opt.str_arg();
  } else {
   assetid=Asset.cur_id();
  }

  if (assetid==null) {
   Shell.err_string="missing assetid argument";
   return null;
  }

  return assetid;
 }

 private static Asset parse_mp4_args(String assetid,Shell.Option[] options) throws Throwable {
  Shell.println("- parsing args");
  String audio_desc,video_desc,time_desc=null;

  Shell.Option opt=Shell.getopt(options,'a');

  if (opt!=null) {
   audio_desc=opt.str_arg();
   } else {
   Shell.err_string="missing audiodesc argument";
   return null;
  }

  opt=Shell.getopt(options,'v');

  if (opt!=null) {
   video_desc=opt.str_arg();
  } else {
   Shell.err_string="missing videodesc argument";
   return null;
  }

  opt=Shell.getopt(options,'t');

  if (opt!=null) {
   time_desc=opt.str_arg();
  } else {
   Shell.err_string="missing timedesc argument";
   return null;
  }

  MP4Builder.StreamsDesc sd=MP4Builder.StreamsDesc.from_args(audio_desc,video_desc);

  if (sd==null) {
   Shell.err_string="invalid audio or video stream description";
   return null;
  }

  Shell.println("  audio name:       "+sd.audio_name());
  Shell.println("  audio quality:    q"+sd.audio_quality());
  Shell.println("  video quality:    q"+sd.video_quality());

  MP4Builder.TimeDesc td=MP4Builder.TimeDesc.from_arg(time_desc);

  if (td==null) {
   Shell.err_string="invalid time description";
   return null;
  }

  Asset asset=new Asset(assetid,sd,td);

  ISMManifest ism=asset.manifest();

  if (ism==null) {
   Shell.err_string="missing Manifest file";
   return null;
  }

  long start_time=td.start_time();
  long duration=td.duration();

  //duration 0 implicates whole movie
  if (duration==0) {
   duration=ism.real_duration();
  }

  long end_time=start_time+duration;

  Shell.println("  start time:       "+start_time+" sec");
  Shell.println("  duration:         "+duration+" sec");

  ISMManifest.AudioQualityLevel aql=ism.get_audio_ql(sd.audio_name(),sd.audio_quality());

  if (aql==null) {
   Shell.err_string="audio stream not present in Manifest file";
   return null;
  }

  ISMManifest.VideoQualityLevel vql=ism.get_video_ql(sd.video_quality());

  if (vql==null) {
   Shell.err_string="video stream not present in Manifest file";
   return null;
  }

  //verify chunk cnt
  ISMManifest.StreamIndex audio_stream=ism.get_stream("audio",sd.audio_name());
  ISMManifest.StreamIndex video_stream=ism.get_stream("video");

  if (audio_stream.chunk_cnt()!=video_stream.chunk_cnt()) {
   Shell.err_string="inconsistent chunk count present in Manifest file for audio and video streams";
   return null;
  }

  return asset;
 }

 // commands
 public static class CMD_DEVICE implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   Device curdev=Device.cur_device();

   curdev.print();

   return true;
  }
 }

 public static class CMD_GRPKEY implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   BigInteger grpkey=Device.get_group_prvkey();

   PaddedPrinter pp=Shell.get_pp();

   pp.println("group prv key: "+grpkey.toString(16));

   return true;
  }
 }

 public static class CMD_GRPCERT implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   BCert.CertificateChain gcert=Device.get_group_cert();

   gcert.print();

   return true;
  }
 }

 public static class CMD_DEVCERT implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   Device curdev=Device.cur_device();

   BCert.Certificate cert=curdev.get_cert();

   cert.print(true);

   return true;
  }
 }

 public static class CMD_DEVCHAIN implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   Device curdev=Device.cur_device();

   BCert.CertificateChain gcert=curdev.get_cert_chain();

   gcert.print();

   return true;
  }
 }

 public static class CMD_MANINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Asset asset=new Asset(assetid);

   ISMManifest ism=asset.manifest();
   ism.print();    

   return true;
  }
 }

 public static class CMD_URLINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Asset asset=new Asset(assetid);
   String s_url=asset.url();

   if (s_url==null) {
    String urlpath=FileCache.url_filename(assetid);
    Shell.err_string="invalid assetid or url file not present ["+urlpath+"]";
    return false;
   }

   Device curdev=Device.cur_device();

   Web.PathInfo pathinfo=CDN.get_pathinfo(curdev.get_serial(),s_url);

   if (pathinfo!=null) {
    pathinfo.print();
   } else {
    Shell.err_string="cannot get url info ["+s_url+"]";
    return false;
   }
   
   return true;
  }
 }

 public static class CMD_GETLICENSE implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Asset asset=new Asset(assetid);
   License license=asset.get_license();

   if (license==null) {
    return false;
   }

   license.print();

   return true;
  }
 }

 public static class CMD_MAKEMP4 implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Asset asset=parse_mp4_args(assetid,options);

   if (asset==null) return false;

   return asset.make_mp4();
  }
 }

 public static class CMD_DOWNMP4 implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Asset asset=parse_mp4_args(assetid,options);

   if (asset==null) return false;

   return asset.download_mp4();
  }
 }

 public static class CMD_ASSETINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Asset asset=new Asset(assetid);

   Asset.Info info=asset.get_info();

   if (info!=null) {
    info.print();     
   } else {
    String infofile=FileCache.info_filename(assetid);
    Shell.err_string="invalid assetid or Info file not present ["+infofile+"]";
    return false;
   }

   return true;
  }
 }

 public static class CMD_MP4DUMP implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String filename=null;
   String boxpath=null;
   boolean verbose=false;

   Shell.Option opt=Shell.getopt(options,'f');

   if (opt!=null) {
    filename=opt.str_arg();
   }

   if (filename==null) {
    Shell.err_string="missing filename argument";
    return false;
   }

   opt=Shell.getopt(options,'b');

   if (opt!=null) {
    boxpath=opt.str_arg();
   }

   opt=Shell.getopt(options,'v');

   if (opt!=null) {
    verbose=true;;
   }

   MP4File mp4=MP4File.from_file(filename);

   if (mp4==null) {
    Shell.err_string="problem loading MP4 file: "+filename;
    return false;
   }

   if (boxpath!=null) {
    MP4File.Box box=mp4.get_box(boxpath);

    if (box==null) {
     Shell.err_string="cannot locate MP4 Box: "+boxpath;
     return false;
    }

    box.print(true);
   } else {
    mp4.print(verbose);
   }
   
   return true;
  }
 }

 public static class CMD_SECRETS implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   String files_desc[]={
    "z1",           "PlayReady private group key",
    "g1",           "PlayReady group certificate",
    "stb-cert.jks", "set-top-box certificate",
    "stb-cert.pwd", "set-top-box certificate password"
   };

   for(int i=0;i<files_desc.length;i+=2) {
    String file=files_desc[i];
    String desc=files_desc[i+1];

    String status="";

    if (Utils.file_exists("secrets"+File.separatorChar+file)) {
     status="OK";
    } else {
     status="MISSING";
    }

    String msg=Utils.padded_string(file,16)+" - "+Utils.padded_string(desc,40)+"["+status+"]";
    Shell.println(msg);    
   }

   return true;
  }
 }

 public static class CMD_BCERTINFO implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String filename=options[0].str_arg();

   if (filename==null) {
    Shell.err_string="invalid filename";
    return false;
   }

   BCert bc=BCert.from_file(filename);

   if (bc!=null) {
    bc.print();
   }

   return true;
  }
 }

 public static class CMD_CHECKWATERMARK implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   int frag_idx=0;
   String quality="0";

   String assetid=parse_assetid(options);

   if (assetid==null) return false;

   Shell.Option opt=Shell.getopt(options,'i');

   if (opt!=null) {
    frag_idx=opt.int_arg();
   }

   opt=Shell.getopt(options,'q');

   if (opt!=null) {
    quality=opt.str_arg();
   }

   Asset asset=new Asset(assetid);
   ISMManifest ism=asset.manifest();   

   ISMManifest.VideoQualityLevel vql=ism.get_video_ql(quality);

   if (vql==null) {
    Shell.err_string="video stream of given quality not present in Manifest file";
    return false;
   }

   return asset.check_watermark(frag_idx,quality);
  }
 }

 public static class CMD_EXTRACTSECRETS implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=1) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String filename=options[0].str_arg();

   if (filename==null) {
    Shell.err_string="invalid filename";
    return false;
   }

   if (!Utils.file_exists(filename)) {
    Shell.err_string="file does not exist ["+filename+"]";
    return false;
   }

   String prvkeyname=BASE_DIR+File.separatorChar+"prv.dat";
   String pubkeyname=BASE_DIR+File.separatorChar+"pub.dat";

   if (!Utils.file_exists(prvkeyname)) {
    Shell.err_string="missing private RSA key file ["+filename+"]";
    return false;
   }

   if (!Utils.file_exists(pubkeyname)) {
    Shell.err_string="missing public RSA key file ["+filename+"]";
    return false;
   }

   RSA.KeyPair keypair=RSA.KeyPair.from_file(prvkeyname,pubkeyname);
   keypair.print("ROOT DEVICE KEY");

   if (!keypair.verify()) {
    Shell.err_string="invalid key pair";
    return false;
   }

   return KeyExtractor.extract_keys(filename,keypair);
  }
 }

 public static class CMD_GENTEST implements CmdHandler {
   static String input_collections[]={
    "PREMIERY VOD+",
    "CANAL+ Premium",
    "HBO"
   };

  static final String PLAYER_PATH = "c:\\Program Files (x86)\\Windows Media Player\\wmplayer.exe";

  static final int FONT_SIZE = 20;

  static int random_int(int max) {
   return ThreadLocalRandom.current().nextInt(max);
  }

  static String gen_collection() {
   return input_collections[random_int(input_collections.length)];
  }

  static void print_text(BufferedImage screen_image,int pos_x,int pos_y,String txt,Color color,int font_size) {
   String lines[]=Utils.tokenize(txt,"\n");

   Graphics2D gfx=screen_image.createGraphics();
   gfx.setPaint(color);

   gfx.setFont(new Font("Courier",Font.BOLD,font_size));

   for(int i=0;i<lines.length;i++) {
    gfx.drawString(lines[i],pos_x,(int)(pos_y+i*(font_size*1.5)));
   }
  }

  static void movie_screenshot(String filename,String txt,String outputfile) {
   String wmp=PLAYER_PATH;
 
   try {
    String curdir=Paths.get(".").toAbsolutePath().normalize().toString();
 
    filename=curdir+File.separatorChar+filename;

    ProcessBuilder pb=new ProcessBuilder(wmp,filename,"/fullscreen");
    Process process=pb.start();

    Thread.currentThread().sleep(4000);

    String format="png";

    String capturefile=outputfile+"."+format;

    Robot robot=new Robot();
   
    Dimension screen_size=Toolkit.getDefaultToolkit().getScreenSize();
    Rectangle capture_rect=new Rectangle(0,0,screen_size.width,screen_size.height);

    BufferedImage screen_image=robot.createScreenCapture(capture_rect);

    print_text(screen_image,40,40,txt,Color.white,28);
  
    print_text(screen_image,190,440,"CANAL+ content security check",Color.yellow,100);
    print_text(screen_image,360,570,"CONTENT NOT SECURE",Color.yellow,100);

    ImageIO.write(screen_image,format,new File(capturefile));

    process.destroy(); 
   } catch(Throwable t) {
    t.printStackTrace();
   }
  }

  public boolean run(Shell.Option[] options) throws Throwable {
   System.err.close();
   System.setErr(System.out);

   PaddedPrinter pp=PaddedPrinter.getInstance();

   //setup fake id 
   Vars.set("SERIAL", "_HELLO_MICROSOFT_");
   Vars.set("MAC",    "012345543210");

   String serial=Vars.get_str("SERIAL");
   String mac=Vars.get_str("MAC");
 
   Shell.out.println("\n\n");

   String tst="";

   tst+="CANAL+ content security check (MS PlayReady protection)\n";
   tst+=" - test date:  "+new Date()+"\n";
   tst+=" - client identity\n";
   tst+="   * SERIAL: "+serial+"\n";
   tst+="   * MAC:    "+mac+"\n";

   String time_desc="04:00+2";
   MP4Builder.TimeDesc td=MP4Builder.TimeDesc.from_arg(time_desc);

   if (td==null) {
    Shell.err_string="invalid time description";
    return false;
   }

   //get the list of collections
   VOD.Collection root=VOD.vod_root();

   //run test for each designated collection
   for(int i=0;i<input_collections.length;i++) {
    String colname=input_collections[i];

    VOD.Collection col=VOD.get_collection_by_name(root,colname);

    if (col==null) {
     Shell.err_string="collection cannot be found: "+colname;
     return false;
    }

    //get collection info
    VOD.AssetBrowseInfo abi=col.get_asset_browse_info();

    //select page in a random fashion
    int page_id=random_int(abi.max_page()-1)+1;

    Vector<VOD.Asset> assets=col.assets_from_page(page_id);

    //select asset in a random fashion
    int asset_idx=random_int(assets.size());

    //selected asset
    VOD.Asset vod_asset=assets.elementAt(asset_idx);
    vod_asset.store();

    String asset_id=vod_asset.id();

    Asset asset=new Asset(asset_id);

    Asset.Info info=asset.get_info();

    if (info!=null) {
     tst+="["+colname+"]\n";
     tst+="- asset: "+asset_id+"\n";
     int total=abi.max_page()*30;
     tst+="  * page id "+page_id+", page pos "+asset_idx+" (~"+total+" assets)"+"\n";

     License license=asset.get_license();

     if (license==null) {
      Shell.err_string="cannot get license for asset: "+asset.id();
      return false;
     }

     tst+="- license request   [OK]\n";

     byte content_key[]=license.get_content_key();
     byte digest_bytes[]=Crypto.SHA256(content_key);

     String digest=Utils.construct_hex_string(digest_bytes);
     String key=Utils.hex_value(content_key[0]&0xff,2);

     for(int j=1;j<content_key.length;j++) {
      key+="xx";
     }

     tst+="  * content key:    "+key+"\n";
     tst+="  * key SHA256:     "+digest+"\n";

     //get audio and video descriptions
     asset=new Asset(asset_id);

     ISMManifest ism=asset.manifest();

     //get audio description
     String audio_desc="audio";

     //for audio stream, the first available audio track is selected
     ISMManifest.StreamIndex audio_stream=ism.get_stream("audio");     
     audio_desc+="="+audio_stream.Name()+".0";

     String video_desc="video";

     //for video stream, the stream with highest quality level is selected
     ISMManifest.StreamIndex video_stream=ism.get_stream("video");
     ISMManifest.QualityLevel highest_quality=video_stream.get_ql(video_stream.ql_cnt()-1);
     video_desc+="="+highest_quality.Index();

     Shell.out.println("audio_desc: "+audio_desc);
     Shell.out.println("video_desc: "+video_desc);

     //String audio_desc="audio=Polski.0";
     //String video_desc="video=4";

     //build target MPEG-4 stream description
     MP4Builder.StreamsDesc sd=MP4Builder.StreamsDesc.from_args(audio_desc,video_desc);

     if (sd==null) {
      Shell.err_string="invalid audio or video stream description";
      return false;
     }

     //create asset with stream information
     asset=new Asset(asset_id,sd,td);

     if (asset.download_mp4()) tst+="- CDN download      [OK]\n";
     if (asset.make_mp4())     tst+="- movie decryption  [OK]\n";

     String screenshot_file="samples/movie_screenshot_"+i;

     tst+="- movie screenshot (<a href=\""+screenshot_file+".png\">open</a>)\n";

     String txt="";
     txt+="collection: "+colname+"\n";
     txt+="asset:      "+asset_id+"\n";
     txt+="time_desc:  "+time_desc+"\n";
     txt+="audio_desc: "+audio_desc+"\n";
     txt+="video_desc: "+video_desc+"\n";

     String mp4_path=FileCache.mp4_filename(asset_id);
     movie_screenshot(mp4_path,txt,screenshot_file);
    }

    Shell.out.println(tst);

    Utils.save_file("cp_content_security_check.txt",tst.getBytes());
   }

   return true;
  }
 }

 public static class CMD_LICCRAWL implements CmdHandler {
  public boolean run(Shell.Option[] options) throws Throwable {
   if (options.length!=2) {
    Shell.err_string="invalid number of arguments";
    return false;
   }

   String colname=options[0].str_arg();
   int cnt=options[1].int_arg();

   PaddedPrinter pp=PaddedPrinter.getInstance();

   //setup fake id 
   Vars.set("SERIAL", "_HELLO_CANALPLUS_");
   Vars.set("MAC",    "012345543210");

   String serial=Vars.get_str("SERIAL");
   String mac=Vars.get_str("MAC");
 
   String tst="";

   Date start_time=new Date();

   tst+="CANAL+ license crawling (MS PlayReady protection)\n";
   tst+=" - test date:  "+start_time+"\n";
   tst+=" - client identity\n";
   tst+="   * SERIAL: "+serial+"\n";
   tst+="   * MAC:    "+mac+"\n";
   Shell.out.println(tst);

   //get the list of collections
   VOD.Collection root=VOD.vod_root();

   VOD.Collection col=VOD.get_collection_by_name(root,colname);

   if (col==null) {
    Shell.err_string="collection cannot be found: "+colname;
    return false;
   }

   //get collection info
   VOD.AssetBrowseInfo abi=col.get_asset_browse_info();

   int total=abi.max_page()*30;

   Shell.out.println(" - collection: "+colname+" (~"+total+" assets)");

   int processed=0;

   for(int page_id=1;page_id<abi.max_page()-1;page_id++) {
    Vector<VOD.Asset> assets=col.assets_from_page(page_id);

    for(int i=0;i<assets.size();i++) {
     VOD.Asset vod_asset=assets.elementAt(i);
     vod_asset.store();

     String asset_id=vod_asset.id();

     Asset asset=new Asset(asset_id);

     Asset.Info info=asset.get_info();

     if (info!=null) {
      PrintWriter org_output=Shell.getOutput();
      Shell.setOutput(new PrintWriter(new ByteArrayOutputStream(),true,Shell.CHARSET));

      License license=null;

      try {
       license=asset.get_license();
      } catch(Throwable t) {}

      Shell.setOutput(org_output);

      if (license!=null) {
       Shell.out.println("    * "+info.atttitle_original);

       byte content_key[]=license.get_content_key();
       byte digest_bytes[]=Crypto.SHA256(content_key);

       String digest=Utils.construct_hex_string(digest_bytes);
       String key=Utils.hex_value(content_key[0]&0xff,2);

       for(int j=1;j<content_key.length;j++) {
        key+="xx";
       }

       Shell.out.println("      content key:    "+key);
       Shell.out.println("      key SHA256:     "+digest);

       processed++;

       if (processed>=cnt) {
        Date end_time=new Date();

        long diff=end_time.getTime()-start_time.getTime();

        Shell.out.println("total time: "+diff/1000+" sec");
   
        return true;
       }
      }
     }
    }
   }

   return true;
  }
 }
}